@@ -11,7 +11,8 @@ from ..argument_utility import check_encoder, EncoderArg
 from ..argument_utility import check_use_gpu, UseGPUArg
 from ..argument_utility import check_augmentation, AugmentationArg
 from ..argument_utility import check_q_func, QFuncArg
-from ..argument_utility import ScalerArg
+from ..argument_utility import ScalerArg, ActionScalerArg
+from ..constants import IMPL_NOT_INITIALIZED_ERROR
 
 
 class SAC(AlgoBase):
@@ -25,27 +26,27 @@ class SAC(AlgoBase):
 
     .. math::
 
-        L(\theta_i) = \mathbb{E}_{s_t, a_t, r_{t+1}, s_{t+1} \sim D,
-                                   a_{t+1} \sim \pi_\phi(\cdot|s_{t+1})} [
-            (y - Q_{\theta_i}(s_t, a_t))^2]
+        L(\theta_i) = \mathbb{E}_{s_t,\, a_t,\, r_{t+1},\, s_{t+1} \sim D,\,
+                                   a_{t+1} \sim \pi_\phi(\cdot|s_{t+1})} \Big[
+            \big(y - Q_{\theta_i}(s_t, a_t)\big)^2\Big]
 
     .. math::
 
-        y = r_{t+1} + \gamma (\min_j Q_{\theta_j}(s_{t+1}, a_{t+1})
-            - \alpha \log (\pi_\phi(a_{t+1}|s_{t+1})))
+        y = r_{t+1} + \gamma \Big(\min_j Q_{\theta_j}(s_{t+1}, a_{t+1})
+            - \alpha \log \big(\pi_\phi(a_{t+1}|s_{t+1})\big)\Big)
 
     .. math::
 
-        J(\phi) = \mathbb{E}_{s_t \sim D, a_t \sim \pi_\phi(\cdot|s_t)}
-            [\alpha \log (\pi_\phi (a_t|s_t))
-              - \min_i Q_{\theta_i}(s_t, \pi_\phi(a_t|s_t))]
+        J(\phi) = \mathbb{E}_{s_t \sim D,\, a_t \sim \pi_\phi(\cdot|s_t)}
+            \Big[\alpha \log (\pi_\phi (a_t|s_t))
+              - \min_i Q_{\theta_i}\big(s_t, \pi_\phi(a_t|s_t)\big)\Big]
 
     The temperature parameter :math:`\alpha` is also automatically adjustable.
 
     .. math::
 
-        J(\alpha) = \mathbb{E}_{s_t \sim D, a_t \sim \pi_\phi\(\cdot|s_t)}
-            [-\alpha (\log (\pi_\phi(a_t|s_t)) + H)]
+        J(\alpha) = \mathbb{E}_{s_t \sim D,\, a_t \sim \pi_\phi(\cdot|s_t)}
+            \bigg[-\alpha \Big(\log \big(\pi_\phi(a_t|s_t)\big) + H\Big)\bigg]
 
     where :math:`H` is a target
     entropy, which is defined as :math:`\dim a`.
@@ -81,12 +82,17 @@ class SAC(AlgoBase):
         n_critics (int): the number of Q functions for ensemble.
         bootstrap (bool): flag to bootstrap Q functions.
         share_encoder (bool): flag to share encoder network.
+        target_reduction_type (str): ensemble reduction method at target value
+            estimation. The available options are
+            ``['min', 'max', 'mean', 'mix', 'none']``.
         update_actor_interval (int): interval to update policy function.
         initial_temperature (float): initial temperature value.
         use_gpu (bool, int or d3rlpy.gpu.Device):
             flag to use GPU, device ID or device.
         scaler (d3rlpy.preprocessing.Scaler or str): preprocessor.
-            The available options are `['pixel', 'min_max', 'standard']`
+            The available options are `['pixel', 'min_max', 'standard']`.
+        action_scaler (d3rlpy.preprocessing.ActionScaler or str):
+            action preprocessor. The available options are ``['min_max']``.
         augmentation (d3rlpy.augmentation.AugmentationPipeline or list(str)):
             augmentation pipeline.
         generator (d3rlpy.algos.base.DataGenerator): dynamic dataset generator
@@ -105,9 +111,10 @@ class SAC(AlgoBase):
     _critic_encoder_factory: EncoderFactory
     _q_func_factory: QFunctionFactory
     _tau: float
-    _n_critics: int
     _bootstrap: bool
+    _n_critics: int
     _share_encoder: bool
+    _target_reduction_type: str
     _update_actor_interval: int
     _initial_temperature: float
     _augmentation: AugmentationPipeline
@@ -134,10 +141,12 @@ class SAC(AlgoBase):
         n_critics: int = 2,
         bootstrap: bool = False,
         share_encoder: bool = False,
+        target_reduction_type: str = "min",
         update_actor_interval: int = 1,
         initial_temperature: float = 1.0,
         use_gpu: UseGPUArg = False,
         scaler: ScalerArg = None,
+        action_scaler: ActionScalerArg = None,
         augmentation: AugmentationArg = None,
         generator: Optional[DataGenerator] = None,
         impl: Optional[SACImpl] = None,
@@ -149,6 +158,7 @@ class SAC(AlgoBase):
             n_steps=n_steps,
             gamma=gamma,
             scaler=scaler,
+            action_scaler=action_scaler,
             generator=generator,
         )
         self._actor_learning_rate = actor_learning_rate
@@ -161,9 +171,10 @@ class SAC(AlgoBase):
         self._critic_encoder_factory = check_encoder(critic_encoder_factory)
         self._q_func_factory = check_q_func(q_func_factory)
         self._tau = tau
-        self._n_critics = n_critics
         self._bootstrap = bootstrap
+        self._n_critics = n_critics
         self._share_encoder = share_encoder
+        self._target_reduction_type = target_reduction_type
         self._update_actor_interval = update_actor_interval
         self._initial_temperature = initial_temperature
         self._augmentation = check_augmentation(augmentation)
@@ -190,9 +201,11 @@ class SAC(AlgoBase):
             n_critics=self._n_critics,
             bootstrap=self._bootstrap,
             share_encoder=self._share_encoder,
+            target_reduction_type=self._target_reduction_type,
             initial_temperature=self._initial_temperature,
             use_gpu=self._use_gpu,
             scaler=self._scaler,
+            action_scaler=self._action_scaler,
             augmentation=self._augmentation,
         )
         self._impl.build()
@@ -200,7 +213,8 @@ class SAC(AlgoBase):
     def update(
         self, epoch: int, total_step: int, batch: TransitionMiniBatch
     ) -> List[Optional[float]]:
-        assert self._impl is not None
+        assert self._impl is not None, IMPL_NOT_INITIALIZED_ERROR
+
         critic_loss = self._impl.update_critic(
             batch.observations,
             batch.actions,
@@ -208,20 +222,29 @@ class SAC(AlgoBase):
             batch.next_observations,
             batch.terminals,
             batch.n_steps,
+            batch.masks,
         )
+
         # delayed policy update
         if total_step % self._update_actor_interval == 0:
             actor_loss = self._impl.update_actor(batch.observations)
-            temp_loss, temp = self._impl.update_temp(batch.observations)
+
+            # lagrangian parameter update for SAC temperature
+            if self._temp_learning_rate > 0:
+                temp_loss, temp = self._impl.update_temp(batch.observations)
+            else:
+                temp_loss, temp = None, None
+
             self._impl.update_critic_target()
             self._impl.update_actor_target()
         else:
             actor_loss = None
             temp_loss = None
             temp = None
+
         return [critic_loss, actor_loss, temp_loss, temp]
 
-    def _get_loss_labels(self) -> List[str]:
+    def get_loss_labels(self) -> List[str]:
         return ["critic_loss", "actor_loss", "temp_loss", "temp"]
 
 
@@ -301,8 +324,8 @@ class DiscreteSAC(AlgoBase):
     _actor_encoder_factory: EncoderFactory
     _critic_encoder_factory: EncoderFactory
     _q_func_factory: QFunctionFactory
-    _n_critics: int
     _bootstrap: bool
+    _n_critics: int
     _share_encoder: bool
     _initial_temperature: float
     _target_update_interval: int
@@ -344,6 +367,7 @@ class DiscreteSAC(AlgoBase):
             n_steps=n_steps,
             gamma=gamma,
             scaler=scaler,
+            action_scaler=None,
             generator=generator,
         )
         self._actor_learning_rate = actor_learning_rate
@@ -355,8 +379,8 @@ class DiscreteSAC(AlgoBase):
         self._actor_encoder_factory = check_encoder(actor_encoder_factory)
         self._critic_encoder_factory = check_encoder(critic_encoder_factory)
         self._q_func_factory = check_q_func(q_func_factory)
-        self._n_critics = n_critics
         self._bootstrap = bootstrap
+        self._n_critics = n_critics
         self._share_encoder = share_encoder
         self._initial_temperature = initial_temperature
         self._target_update_interval = target_update_interval
@@ -393,7 +417,8 @@ class DiscreteSAC(AlgoBase):
     def update(
         self, epoch: int, total_step: int, batch: TransitionMiniBatch
     ) -> List[Optional[float]]:
-        assert self._impl is not None
+        assert self._impl is not None, IMPL_NOT_INITIALIZED_ERROR
+
         critic_loss = self._impl.update_critic(
             batch.observations,
             batch.actions,
@@ -401,12 +426,21 @@ class DiscreteSAC(AlgoBase):
             batch.next_observations,
             batch.terminals,
             batch.n_steps,
+            batch.masks,
         )
+
         actor_loss = self._impl.update_actor(batch.observations)
-        temp_loss, temp = self._impl.update_temp(batch.observations)
+
+        # lagrangian parameter update for SAC temeprature
+        if self._temp_learning_rate > 0:
+            temp_loss, temp = self._impl.update_temp(batch.observations)
+        else:
+            temp_loss, temp = None, None
+
         if total_step % self._target_update_interval == 0:
             self._impl.update_target()
+
         return [critic_loss, actor_loss, temp_loss, temp]
 
-    def _get_loss_labels(self) -> List[str]:
+    def get_loss_labels(self) -> List[str]:
         return ["critic_loss", "actor_loss", "temp_loss", "temp"]
