@@ -22,12 +22,11 @@ from ...models.optimizers import OptimizerFactory
 from ...models.encoders import EncoderFactory
 from ...models.q_functions import QFunctionFactory
 from ...gpu import Device
-from ...preprocessing import Scaler
+from ...preprocessing import Scaler, ActionScaler
 from ...augmentation import AugmentationPipeline
 from ...torch_utility import (
     torch_api,
     train_api,
-    eval_api,
     hard_sync,
     augmentation_api,
 )
@@ -64,9 +63,11 @@ class SACImpl(DDPGBaseImpl):
         n_critics: int,
         bootstrap: bool,
         share_encoder: bool,
+        target_reduction_type: str,
         initial_temperature: float,
         use_gpu: Optional[Device],
         scaler: Optional[Scaler],
+        action_scaler: Optional[ActionScaler],
         augmentation: AugmentationPipeline,
     ):
         super().__init__(
@@ -84,8 +85,10 @@ class SACImpl(DDPGBaseImpl):
             n_critics=n_critics,
             bootstrap=bootstrap,
             share_encoder=share_encoder,
+            target_reduction_type=target_reduction_type,
             use_gpu=use_gpu,
             scaler=scaler,
+            action_scaler=action_scaler,
             augmentation=augmentation,
         )
         self._temp_learning_rate = temp_learning_rate
@@ -158,7 +161,13 @@ class SACImpl(DDPGBaseImpl):
         with torch.no_grad():
             action, log_prob = self._policy.sample_with_log_prob(x)
             entropy = self._log_temp().exp() * log_prob
-            return self._targ_q_func.compute_target(x, action) - entropy
+            target = self._targ_q_func.compute_target(
+                x, action, reduction=self._target_reduction_type
+            )
+            if self._target_reduction_type == "none":
+                return target - entropy.view(1, -1, 1)
+            else:
+                return target - entropy
 
 
 class DiscreteSACImpl(DiscreteQFunctionMixin, TorchImplBase):
@@ -208,7 +217,9 @@ class DiscreteSACImpl(DiscreteQFunctionMixin, TorchImplBase):
         scaler: Optional[Scaler],
         augmentation: AugmentationPipeline,
     ):
-        super().__init__(observation_shape, action_size, scaler, augmentation)
+        super().__init__(
+            observation_shape, action_size, scaler, None, augmentation
+        )
         self._actor_learning_rate = actor_learning_rate
         self._critic_learning_rate = critic_learning_rate
         self._temp_learning_rate = temp_learning_rate
@@ -302,6 +313,7 @@ class DiscreteSACImpl(DiscreteQFunctionMixin, TorchImplBase):
         obs_tpn: torch.Tensor,
         ter_tpn: torch.Tensor,
         n_steps: torch.Tensor,
+        masks: torch.Tensor,
     ) -> np.ndarray:
         assert self._critic_optim is not None
 
@@ -311,7 +323,7 @@ class DiscreteSACImpl(DiscreteQFunctionMixin, TorchImplBase):
         q_tpn *= 1.0 - ter_tpn
 
         loss = self.compute_critic_loss(
-            obs_t, act_t.long(), rew_tpn, q_tpn, n_steps
+            obs_t, act_t.long(), rew_tpn, q_tpn, n_steps, masks
         )
 
         loss.backward()
@@ -344,8 +356,11 @@ class DiscreteSACImpl(DiscreteQFunctionMixin, TorchImplBase):
         rew_tpn: torch.Tensor,
         q_tpn: torch.Tensor,
         n_steps: torch.Tensor,
+        masks: torch.Tensor,
     ) -> torch.Tensor:
-        return self._compute_critic_loss(obs_t, act_t, rew_tpn, q_tpn, n_steps)
+        return self._compute_critic_loss(
+            obs_t, act_t, rew_tpn, q_tpn, n_steps, masks
+        )
 
     def _compute_critic_loss(
         self,
@@ -354,10 +369,11 @@ class DiscreteSACImpl(DiscreteQFunctionMixin, TorchImplBase):
         rew_tpn: torch.Tensor,
         q_tpn: torch.Tensor,
         n_steps: torch.Tensor,
+        masks: torch.Tensor,
     ) -> torch.Tensor:
         assert self._q_func is not None
         return self._q_func.compute_error(
-            obs_t, act_t, rew_tpn, q_tpn, self._gamma ** n_steps
+            obs_t, act_t, rew_tpn, q_tpn, self._gamma ** n_steps, masks=masks
         )
 
     @train_api
@@ -423,12 +439,9 @@ class DiscreteSACImpl(DiscreteQFunctionMixin, TorchImplBase):
         assert self._policy is not None
         return self._policy.best_action(x)
 
-    @eval_api
-    @torch_api(scaler_targets=["x"])
-    def sample_action(self, x: torch.Tensor) -> np.ndarray:
+    def _sample_action(self, x: torch.Tensor) -> torch.Tensor:
         assert self._policy is not None
-        with torch.no_grad():
-            return self._policy.sample(x).cpu().detach().numpy()
+        return self._policy.sample(x)
 
     def update_target(self) -> None:
         assert self._q_func is not None
